import std;
import algebra;
import harness;

const unsigned int kViewMatrix        = 0;
const unsigned int kPerspectiveMatrix = 4;
const unsigned int kScales            = 8;
const unsigned int kMeshNumVertices   = 36;
const unsigned int kMeshVertices      = 9;
const unsigned int kMeshRenderTag     = 117;
const unsigned int kTextureUploadTag  = 226;

// Euler angles for the current rotation of the mesh.
const unsigned int kEuler             = 1021; // VU mem address where we store the mesh rotation.

// The status variable where we store whether we have uploaded the texture yet.
const unsigned int kTextureDone       = 1022; // VU mem address.


// First build a model matrix from euler angles, then concatenate the matrices to go 
// from model space to screen space (model * view * perspective).
void BuildTransformMatrix(float4 result[4], float4 vuMem[], const uint modelEulerOfs, const uint viewMatrixOfs, const uint perspectiveMatrixOfs)
{
  float4 euler = vuMem[modelEulerOfs];
  float4 modelMatrix[4];
  MatrixBuildRotateXYZ(modelMatrix, euler);
  
  float4 viewMatrix[4] = vuMem + viewMatrixOfs;

  MatrixMultiply(result, modelMatrix, viewMatrix);

  float4 perspectiveMatrix[4] = vuMem + perspectiveMatrixOfs;
  MatrixMultiply(result, result, perspectiveMatrix);
}

// Draw a textured mesh.
//
// The vertex array is assumed to be:
//   Pre-transformed:  { texCoord, position }
//   Post-transformed: { texCoord, color, position }
//
// transformMatrix     - The final transformation matrix that should be applied to get the vertices to post-transformed state.
// scales              - ToDo: Should be removed and baked into the transformation address or generated on the fly.
void MeshDraw(float4 transformMatrix[4], float4 scales, const uint numVertices, float4 vertices[], float4 renderTag[])
{
  // Loop through transforming each model vertex, storing transformed verts into GIFtag.
  float4 src[]     = vertices;
  float4 srcEnd[]  = vertices + numVertices * 3;
  float4 dst[]     = renderTag + 1;
  
  do
  {
    // Load up the model space vertex data into tmp registers.
    float4 texCoord = src[0];
    float4 vertex   = src[2];

    // Transform the vertex.
    vertex = MatrixMultiplyVertex(transformMatrix, vertex);
    ViewportTransformVT(vertex, texCoord, scales);
    
    // Save the result into the post-transform vertex array.
    dst[0] = texCoord;
    dst[2] = vertex;
    
    // Move on to the next vertex.
    src += 3;
    dst += 3;
  }
  while (src != srcEnd);

  // Send the vertices to the GS.
  xgkick(renderTag);
}

void main(float4 vuMem[])
{
  // Check if we should upload the texture.
  int iTextureDone = int(vuMem[kTextureDone].x);
  if (iTextureDone == 0)
  {
    // Upload the texture.
    xgkick(vuMem + kTextureUploadTag);

    // Now that it's uploaded, skip this code the following frames.
    vuMem[kTextureDone].x = float(1);
  }

  // Poll input.
  int padInfo = PadPoll(vuMem);

  if (PadPressed(kPadR1, padInfo))
  {
    float4 cameraZoom = vuMem[kViewMatrix + 3];
    cameraZoom.z -= 1.0f;
    vuMem[kViewMatrix + 3] = cameraZoom;
  }
  else if (PadPressed(kPadR2, padInfo))
  {
    float4 cameraZoom = vuMem[kViewMatrix + 3];
    cameraZoom.z += 1.0f;
    vuMem[kViewMatrix + 3] = cameraZoom;
  }
  else if (PadPressed(kPadLeft, padInfo))
  {
    float4 euler = vuMem[kEuler];
    euler.x += 0.1f;
    vuMem[kEuler] = euler;
  }
  else if (PadPressed(kPadRight, padInfo))
  {
    float4 euler = vuMem[kEuler];
    euler.x -= 0.1f;
    vuMem[kEuler] = euler;
  }
  else if (PadPressed(kPadUp, padInfo))
  {
    float4 euler = vuMem[kEuler];
    euler.y += 0.1f;
    vuMem[kEuler] = euler;
  }
  else if (PadPressed(kPadDown, padInfo))
  {
    float4 euler = vuMem[kEuler];
    euler.y -= 0.1f;
    vuMem[kEuler] = euler;
  }
  else if (PadPressed(kPadTriangle, padInfo))
  {
    float4 euler = vuMem[kEuler];
    euler.z += 0.1f;
    vuMem[kEuler] = euler;
  }
  else if (PadPressed(kPadCross, padInfo))
  {
    float4 euler = vuMem[kEuler];
    euler.z -= 0.1f;
    vuMem[kEuler] = euler;
  }
  
  // Draw mesh.
  float4 transformMatrix[4];
  BuildTransformMatrix(transformMatrix, vuMem, kEuler, kViewMatrix, kPerspectiveMatrix);
  MeshDraw(transformMatrix, vuMem[kScales], kMeshNumVertices, vuMem + kMeshVertices, vuMem + kMeshRenderTag);
}
